/*
 * Copyright 1992 Oliver Jones
 * Portions Copyright 1989 Massachusetts Institute of Technology
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without
 * fee, provided that the above copyright notice appear in all copies
 * and that both that copyright notice and this permission notice
 * appear in supporting documentation, and that the name of M.I.T. not
 * be used in advertising or publicity pertaining to distribution of
 * the software without specific, written prior permission.  No
 * representation are made by anyone about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 */

#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/AsciiText.h>
#include <setjmp.h>
#include <pwd.h>


#define ASSERT(expr) if (!(expr)) abort();

typedef enum Votestate {
  Unopened, Failure, Opened, Voted, Closed
} Votestate;

typedef struct Choice {
  char *     choicename;   /* display string for choice */
  int        tally;        /* number voting for this choice */
} Choice;

typedef struct Vote {
  char *        votername;
  Choice *      choice;
  Votestate     state;
  Display *     dpy;
  Widget        top;
  Widget        text;
  Widget        comment;
  char 		*commentText;
} Vote;

typedef struct Election {
  int             votesmax;    /* max number of votes (size of array) */
  int             votescur;    /* curr number of votes */
  Vote         ** votes;       /* pointer to array of votes */
  int             choicesmax;  /* max number of choices */
  int             choicescur;  /* curr number of choices */
  Choice       ** choices;     /* pointer to array of choices */
  char          * question;    /* question to ask */
  Boolean         secret;      /* true for secret vote */
  XtAppContext    app;         /* application context */
  int             opencount;   /* open displays */
  char *          argv0;       /* program name */
  XtIntervalId    reportReq;   /* deferred report request */
  char         report[BUFSIZ]; /* name of report file */
  char 		*user;		/* person requesting vote */
} Election;

typedef struct Voterfile {
  Election * e;
  FILE     * fp;
} Voterfile;
  
/* global data structure */
Election election;

jmp_buf XIOrecover;

/* functions */
void Close ();
static void CommentAdded ();
void inputVoterfile ();
static void nullTimerProc ();
char * ParseDisplay ();
char * ParseVotername ();
static void doReport ();
static void requestReport ();

static XtActionsRec actions [] = {
  {"Close",         Close},
  {"comment-added", CommentAdded}, 
};

static void
destroyDisplay (dpy)
     Display * dpy;
{
  int i;
  int count = 0;
  Election * e = &election;

  for (i = 0; i < e->votescur; i++ ){
    Vote * vote = e->votes[i];
    if (vote->dpy == dpy) {
      /* this is the display to close */
      XtDestroyWidget (vote->top);
      XtCloseDisplay (dpy);
      vote->state = Closed;
      vote->dpy   = 0;
      vote->top   = 0;
      vote->text  = 0;
      vote->comment = 0;
    }
    if (vote->dpy) count++;
  }

  /* did we just close the last display connection? */
  if (count == 0) {
    /* terminate */
    unlink (e->report);
    doReport (e, stdout);
    XtFree(e->user);
    for (i = 0; i < e->votescur; i++ ) XtFree(e->votes[i]->commentText);
    exit (0);
  }
  e->opencount = count;
  requestReport(e, 500L);
}

/*ARGSUSED*/
void Close (w, ev, params, num_params)
     Widget w;
     XEvent * ev;
     String params[];
     Cardinal * num_params;
{
  destroyDisplay (XtDisplay(w));
}


XIOhandler(dpy)
     Display * dpy;
{
  destroyDisplay (dpy);
  longjmp (XIOrecover, 1);
}


static void
initElection (e)
     Election * e;
{
  int nalloc;
  FILE * fp;

  nalloc = 5;
  e->votes = (Vote **) XtCalloc (sizeof(Vote *), nalloc);
  e->votesmax = nalloc;
  e->votescur = 0;

  nalloc = 3;
  e->choices = (Choice **) XtCalloc (sizeof(Choice *), nalloc);
  e->choicesmax = nalloc;
  e->choicescur = 0;

  e->opencount = 0;
  e->secret = False;
  e->question = 0;

	{
	/* ripped from XtGetRootDirName; portability unclear */
	extern char *getenv();
#ifndef X_NOT_POSIX
     uid_t uid;
#else
     int uid;
     extern int getuid();
#ifndef SYSV386
     extern struct passwd *getpwuid(), *getpwnam();
#endif
#endif
	struct passwd *pw;
	static char *ptr = NULL;

        if (ptr = getenv("USER")) pw = getpwnam(ptr);
            else {
               uid = getuid();
               pw = getpwuid(uid);
            }
       if (pw) ptr = pw->pw_name;

  	e->user = ptr ? XtNewString(ptr) : XtNewString("unknown");
	}

  sprintf (e->report, "/tmp/votereport%d", getpid());
  fp = fopen (e->report, "w");
  fclose (fp);
}



static void
argQuestion (e, a)
     Election * e;
     char *   a;
{
  if (!e->question) {
    e->question = (char *) malloc (BUFSIZ);
    ASSERT(e->question);
    e->question[0] = '\0';
  } else {
    if (1 + strlen(e->question) < BUFSIZ) 
      strcat (e->question, " ");
  }
  if (strlen(a) + strlen(e->question) < BUFSIZ) 
    strcat (e->question, a);
}

static void
argAnswers (e, a)
     Election * e;
     char *   a;
{
  char * p = a;
  char * l;
  int len;
  Choice * c;

  while (*p) {
    l = index (p, '|');
    if (!l) l = &p[strlen(p)];
    if ((len = l-p) > 0) {
      c = (Choice *) XtMalloc (sizeof(Choice));
      ASSERT(c);

      c->tally = 0;

      /* store the string for the choice name */
      c->choicename = (char *) malloc (len + 1);
      ASSERT (c->choicename);
      strncpy (c->choicename, p, len);
      c->choicename[len] = '\0';

      /* make sure we have enough choices */
      if (e->choicescur >= e->choicesmax) {
	e->choicesmax *= 2;
	e->choices =  (Choice **) XtRealloc (e->choices,
	                 sizeof(Choice *) * e->choicesmax);
	ASSERT(e->choices);
      }

      /* store this choice in the choices array */
      e->choices[e->choicescur] = c;
      e->choicescur++;

    }
    p = l;
    if (*p == '|') p++;
  }
}


static void
argVoterfile (e, a)
     Election * e;
     char *   a;
{

  Voterfile * vf = (Voterfile *) XtMalloc(sizeof(Voterfile));

  ASSERT(vf);
  vf->e = e;

  /* Open the named pipe for reading */
  vf->fp = fopen (a, "r");
  if (!vf->fp) {
    fprintf (stderr, "%s: %s not found\n", 
	     e->argv0, a);
    exit(1);
  }
  /* register the input handler  */
  (void) XtAppAddInput (election.app, fileno(vf->fp),
			   XtInputReadMask, inputVoterfile, vf);
}

static void
argsElection (e, argc, argv)
     Election * e;
     int        argc;
     char     * argv[];
{
  char ** argp = &argv[1];
  char ** argl = &argv[argc];
  char * arg;
  Boolean foundquestion = False;
  Boolean foundvoterfile = False;
  Boolean foundanswers = False;
  
  while ( argp < argl) {
    arg = *argp++;
    if (arg[0] == '-') {
      if (arg[1] == 's') e->secret = True;
      if (arg[1] == 'v') {
	if (argp < argl) arg = *argp++;
	argVoterfile (e, arg);
	foundvoterfile = True;
      }
      if (arg[1] == 'a') {
	if (argp < argl) arg = *argp++;
	argAnswers (e, arg);
	foundanswers = True;
      }
    }
    else {
      argQuestion (e, arg);
      foundquestion = True;
    }
  }
  if (!foundquestion) argQuestion (e, "What is your vote?");
  if (!foundanswers) argAnswers  (e, "Yes|No|Abstain");
  if (!foundvoterfile) argVoterfile (e, "/usr/local/lib/vote/default.voters");

}

main (argc, argv)
  int    argc;
  char  *argv[];
{

  Election * e = &election;

  /* initialize the application data structure */
  initElection (e);
  /* initialize the application name */
  e->argv0 = argv[0];

  /* Initialize the intrinsics */
  XtToolkitInitialize ();
  e->app = XtCreateApplicationContext ();
  /* set up the Close action */
  XtAppAddActions(e->app, actions, XtNumber(actions));

  /* interpret command line args */
  argsElection (e, argc, argv);

  /* error recovery, and main loop */
  (void) setjmp (XIOrecover);
  XSetIOErrorHandler (XIOhandler);
  XtAppMainLoop(election.app);
}

/* ARGSUSED */
void
quitCallback (w, client, call)
     Widget       w;
     XtPointer client, call;
{
  destroyDisplay (XtDisplay(w)); /* go down the hard way */
}

/* ARGSUSED */
void
callbackVote (w, c, state)
     Widget       w;
     Choice *     c;
     int          state;
{
  int i;
  Display * dpy = XtDisplay(w);
  Election * e = &election;
  
  for (i = 0; i < e->votescur; i++ ){
    Vote * vote = e->votes[i];
    if (vote->dpy == dpy) {
      if (state == 0 && vote->choice == c) {
	vote->choice = 0;
	vote->state = Opened;
      }
      else {
	vote->choice = c;
	vote->state = Voted;
      }
      requestReport (e, 500L);
      break;
    }
  }
}

static void
wmDeleteWindow (w)
     Widget w;
{

  Atom wm_delete_window;

  wm_delete_window = XInternAtom (XtDisplay(w), 
				  "WM_DELETE_WINDOW", False);
  (void) XSetWMProtocols  (XtDisplay(w), 
			   XtWindow(w),
			   &wm_delete_window, 1);
  
  XtOverrideTranslations  (w,
			   XtParseTranslationTable
			   ("<Message>WM_PROTOCOLS: Close()"));
}


void
createBallot (e, vp)
     Election * e;
     Vote * vp;
{
  int i;
  Arg args[4];
  char title[BUFSIZ];
  Widget pane, box;

  strcpy (title, "Vote - ");
  strcat (title, vp->votername);

  XtSetArg (args[0], XtNtitle, (XtArgVal) title);

  vp->top = XtAppCreateShell (NULL, "Vote",
			      applicationShellWidgetClass,
			      vp->dpy, args, 1L);
  
  pane = XtCreateManagedWidget("pane", panedWidgetClass,
			       vp->top, NULL, 0);

  XtSetArg (args[0], XtNshowGrip, (XtArgVal) False);
  XtSetArg (args[1], XtNlabel, (XtArgVal) e->question);
  
  XtCreateManagedWidget("question", labelWidgetClass,
			pane, args, 2L);

  box = XtCreateManagedWidget("answers", boxWidgetClass,
			pane, args, 1L);

  XtSetArg (args[1], XtNstring, (XtArgVal) e->report);
  
  vp->text = XtCreateManagedWidget ("results", asciiTextWidgetClass,
				    pane, args, 2L);

  XtSetArg (args[1], XtNskipAdjust, (XtArgVal) True);
  XtSetArg (args[2], XtNeditType, (XtArgVal) XawtextEdit);
  vp->comment = XtCreateManagedWidget ("comment", asciiTextWidgetClass,
				    pane, args, 3L);
  XtOverrideTranslations(vp->comment, 
		XtParseTranslationTable("<Key>Return: comment-added()"));
  vp->commentText = (char*)NULL;

  XtAddCallback(XtCreateManagedWidget("quit", commandWidgetClass,
			pane, args, 2L), XtNcallback, quitCallback, (XtPointer)e);

  for (i = 0; i < e->choicescur; i++) {

    Choice * c = e->choices[i];
    Widget w;
    int nargs = 2;
    if (i != 0) nargs = 3;

    XtSetArg (args[1], XtNradioData, (XtArgVal) c);

    w = XtCreateManagedWidget(c->choicename,
			      toggleWidgetClass,
			      box, args, nargs);

    XtAddCallback (w, XtNcallback, callbackVote,
		   (XtPointer) c);
    if (i==0) 
      XtSetArg (args[2], XtNradioGroup, (XtArgVal) w);

  }
  
  XtRealizeWidget (vp->top);
  
  wmDeleteWindow (vp->top);
}

void
inputVoterfile (vf, fd, inputid)
     Voterfile *  vf;
     int * fd;
     XtInputId * inputid;
{
  char buf [BUFSIZ];
  char * dpyname;
  char * votername;
  Election * e = vf->e;
  Vote ** votep;
  Vote * vote;
  Vote * vp = 0;
  int i;

  ASSERT(fileno(vf->fp) == *fd);

  /* read, handle possible error or EOF condition */
  if (!fgets (buf, BUFSIZ, vf->fp))  {
    XtRemoveInput(*inputid);
    XtFree (vf);
    return;
  }

  buf[strlen(buf)-1] = '\0'; /* chop line */
  if (strlen(buf) == 0) return;
  if (buf[0] == '#') return;
  dpyname = ParseDisplay (buf);
  votername = ParseVotername (buf);
  
  votep = &e->votes[0];
  for (i = 0; i < e->votescur; i++) {
    vote = *votep++;
    if (0 == strcmp(votername, vote->votername)) {
      vp = vote;
      XtFree (votername);
      XtFree (dpyname);
    }
  }  

  if (!vp) {  /* ensure we don't have a display already */
    Display *dpy;
    int argc = 3;
    char * argv[4]; 
    argv[0]= e->argv0;
    argv[1]= "-geometry";	/* If the X display is of the sort that 
				** times out on an action, then leaving this
				** window up on a screen, pending placement
				** by a window manager, can cause the
				** remote user's session to be lost. make
				** a (naive) attempt to locate the window
				** to avoid this; quasi-center it for the
				** size of some displays to make it more
				** noticeable.
				** This option should really just be passed
				** along from the real command-line, which
				** the program is essentially ignoring.
				*/
    argv[2]= "+500+400";
    argv[3]= (char*)NULL;
    vp = (Vote *) XtCalloc (1, sizeof(Vote));
    ASSERT (vp);

    vp->state = Unopened;
    vp->votername = votername;

    /* make sure we have enough entries in the voter list */
    if (e->votescur >= e->votesmax) {
      e->votesmax *= 2;
      e->votes =  (Vote **) XtRealloc 
	(e->votes, sizeof(Vote *) * e->votesmax);
      ASSERT(e->votes);
    }
    
    /* store this vote in the votes array */
    e->votes[e->votescur] = vp;
    e->votescur++;

    dpy= XtOpenDisplay (e->app, dpyname,
			NULL, "Vote", 
			NULL, 0L,
			&argc, &argv);

    if (!dpy) {
      vp->state=Failure;
      return;
    }
	XBell(dpy,0);
    
    vp->dpy = dpy;
    vp->state = Opened;

    /* set up the app structure */
    createBallot (e, vp);
  }

  /* this works around a bug */
  (void) XtAppAddTimeOut(e->app, 1L, nullTimerProc, e);
  requestReport(&election, 500L);
}

static void
nullTimerProc ()
{
  /* EMPTY */
  /* This null timer proc is used to work around     */
  /* a bug which prevents the main loop from seeing  */
  /* events from a new display the first time around */
}


char *
ParseDisplay (buf)
     char * buf;
{
  char * res;
  char * l = (char * ) strchr(buf,' ');
  int len;

  if (l) {
    len = l - buf;
    res = XtCalloc(1,len+1);			/* items not freed if run to end */
    ASSERT(res);
    if(len) strncpy (res,buf,len);
  } else {
    res = XtCalloc(1,strlen(buf)+1);
    ASSERT(res);
    strcpy(res, buf);
  }
  return res;
}		 
  

char *
ParseVotername (buf)
     char * buf;
{
  char * res;
  char * l = (char * ) strchr(buf,' ');
  int len;

  if (l)  l++;
  else l = buf;

  len = strlen(l);
  res = XtCalloc(1,len+1);
  ASSERT(res);
  strcpy(res, l);
  return res;
}		 
  
static void
doReport (e, fp)
     Election * e;
     FILE * fp;
{

  int i;
  int count = 0;
  double percent;

  if (e->votescur <= 0) {
    fprintf (fp, "Error--no voters\n");
    return;
  }

  for (i=0; i < e->choicescur; i++) {
    e->choices[i]->tally = 0;
  }

  for (i = 0; i < e->votescur; i++ ){
    Vote * vote = e->votes[i];
    if (vote->choice)  {
      vote->choice->tally++;
      count++;
    }
  }


  fprintf (fp, "Results of query from %s: \"%s\"\n\n", e->user, e->question);

  percent = (100.0 * count) / election.votescur;
  fprintf (fp, "  %ld of %ld (%3.1f%%) votes have been cast.\n\n",
	   count, election.votescur, percent);
  fprintf (fp, "     Total   Percent   Vote\n");
  fprintf (fp, "     -----   -------   ----\n");
  for (i=0; i < election.choicescur; i++) {
    Choice * choice = e->choices[i];
    if (choice->tally) {
      percent = (100.0 * choice->tally) / count;
      fprintf (fp, "     %5ld    %5.1f%%   %s\n", 
	       choice->tally, percent, choice->choicename);
    }
    else {
      fprintf (fp, "     %5d             %s\n", 
	       choice->tally, choice->choicename);
    }
  }

  fprintf (fp, "\n");
  fprintf (fp,"          Vote Status  Voter \tComment\n");
  fprintf (fp,"          -----------  ----- \t-------\n");
  for (i = 0; i < election.votescur; i++ ){
    Vote * vote = e->votes[i];
	char *comment = (char*)e->votes[i]->commentText;
	if (e->votes[i]->comment)
	  {
	  Arg arg[1];
	  XtSetArg(arg[0], XtNstring,(XtArgVal)&comment);
	  XtGetValues(e->votes[i]->comment, arg, 1);
          XtFree(e->votes[i]->commentText);
	  e->votes[i]->commentText = comment ? XtNewString(comment) : NULL;
	  }
	if ((char*)NULL == comment) comment = "<none>";

    if (vote->state == Failure) 
      fprintf (fp,"<Cannot Open Display>  %s \t%s\n", vote->votername,comment);
    else if (!vote->choice)
      fprintf (fp,"          <Not Voted>  %s \t%s\n", vote->votername,comment);
    else if (e->secret)
      fprintf (fp,"              <Voted>  %s \t%s\n", vote->votername,comment);
    else
      fprintf (fp,"%21s  %s \t%s\n", vote->choice->choicename,vote->votername,comment);
  }
}

/*ARGSUSED1 */
static void
showReport (e, t)
     Election * e;
     XtIntervalId * t;
{
  Arg arg;
  int i;
  FILE * fp = fopen (e->report, "w");

  doReport (e, fp);
  fclose(fp);

  XtSetArg (arg, XtNstring, (XtArgVal) e->report);
  for (i = 0; i < election.votescur; i++ ){
    Widget text = e->votes[i]->text;
    if (text) XtSetValues (text, arg, 1L);
  }

  e->reportReq = 0;
}

static void
requestReport (e, time)
     Election * e;
     long int time;
{
  if (!e->reportReq)
    e->reportReq = XtAppAddTimeOut(e->app, time, 
				   showReport, e);
}

static void CommentAdded (w, ev, params, num_params)
     Widget w;
     XEvent * ev;
     String params[];
     Cardinal * num_params;
{
      Election * e = &election;
      requestReport (e, 500L);
}
